"""
This is a automatically generated file from Psi4's density functional metadata.
Psi4 Version: 1.4a2.dev215

File Authors: QCElemental Authors
"""


data_blob = {
    "version": "1.4a2.dev215",
    "functionals": {
        "teter93": {
            "ansatz": 0,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "zlp": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "ksdt": {
            "ansatz": 0,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "hcth407p": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "hcthp76": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "hcthp14": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "b97-gga1": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "kt2": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "th1": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "th2": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "th3": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "th4": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "hcth": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "hcth93": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "hcth120": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "hcth147": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "hcth407": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "edf1": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "xlyp": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "pbe1w": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "mpwlyp1w": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "pbelyp1w": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "mohlyp": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "mohlyp2": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "th-fl": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "th-fc": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "th-fcfo": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "th-fco": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "vv10": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": True,
            "deriv": 1,
        },
        "b97-1p": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "b3pw91": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "b3lyp": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "b3p86": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "o3lyp": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "mpw1k": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "pbeh": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "pbe0": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "b97-0": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "b97-1": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "b97-2": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "x3lyp": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "b1wc": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "b97-k": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "b97-3": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "mpw3pw": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "b1lyp": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "b1pw91": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "mpw1pw91": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "mpw1pw": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "mpw3lyp": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "sb98-1a": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "sb98-1b": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "sb98-1c": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "sb98-2a": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "sb98-2b": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "sb98-2c": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "hse03": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": True,
            "nlc": False,
            "deriv": 1,
        },
        "hse06": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": True,
            "nlc": False,
            "deriv": 1,
        },
        "hjs-pbe": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": True,
            "nlc": False,
            "deriv": 1,
        },
        "hjs-pbesol": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": True,
            "nlc": False,
            "deriv": 1,
        },
        "hjs-pbe-sol": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": True,
            "nlc": False,
            "deriv": 1,
        },
        "hjs-b88": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": True,
            "nlc": False,
            "deriv": 1,
        },
        "hjs-b97x": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": True,
            "nlc": False,
            "deriv": 1,
        },
        "cam-b3lyp": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": True,
            "nlc": False,
            "deriv": 1,
        },
        "tuned-cam-b3lyp": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": True,
            "nlc": False,
            "deriv": 1,
        },
        "bhandh": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "bhhlyp": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "bhandhlyp": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "mb3lyp-rc04": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "mpwlyp1m": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "revb3lyp": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "pbe0-13": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "b3lyps": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "wb97": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": True,
            "nlc": False,
            "deriv": 1,
        },
        "wb97x": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": True,
            "nlc": False,
            "deriv": 1,
        },
        "lrc-wpbeh": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": True,
            "nlc": False,
            "deriv": 1,
        },
        "wb97xv": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": True,
            "nlc": True,
            "deriv": 1,
        },
        "wb97x-v": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": True,
            "nlc": True,
            "deriv": 1,
        },
        "lc-vv10": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": True,
            "nlc": True,
            "deriv": 1,
        },
        "hpbeint": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "lrc-wpbe": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": True,
            "nlc": False,
            "deriv": 1,
        },
        "b3lyp5": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "edf2": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "cap0": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "b1b95": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "b88b95": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "b86b95": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "pw86b95": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "bb1k": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "mpw1b95": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "mpwb1k": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "x1b95": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "xb1k": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "pw6b95": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "pwb6k": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "tpss0": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "tpssh": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "revtpssh": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "wb97mv": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": True,
            "nlc": True,
            "deriv": 1,
        },
        "wb97m-v": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": True,
            "nlc": True,
            "deriv": 1,
        },
        "tpsslyp1w": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "b97m-v": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": True,
            "deriv": 1,
        },
        "b5050lyp": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "svwn": {
            "ansatz": 0,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "blyp": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "bop": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "b86bpbe": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "pw86pbe": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "pbe": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "rpbe": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "revpbe": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "n12": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "pw91": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "mpwpw": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "bp86": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "ft97": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "sogga": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "sogga11": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "gam": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "oppbe": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "pbeop": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "op-pbe": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "m06l": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "m06-l": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "revm06l": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "revm06-l": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "m11l": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "m11-l": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "mn12l": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "mn12-l": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "mn15l": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "mn15-l": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "mgga-ms0": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "mgga_ms0": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "mgga-ms1": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "mgga_ms1": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "mgga-ms2": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "mgga_ms2": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "mgga-mvs": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "mgga_mvs": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "tpss": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "revtpss": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "pkzb": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "vsxc": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "scan": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "revscan": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "lc-wpbe0": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": True,
            "nlc": False,
            "deriv": 1,
        },
        "wpbe0": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": True,
            "nlc": False,
            "deriv": 1,
        },
        "lc-wpbe": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": True,
            "nlc": False,
            "deriv": 1,
        },
        "lcwpbe": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": True,
            "nlc": False,
            "deriv": 1,
        },
        "wpbe": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": False,
            "c_lrc": False,
            "x_lrc": True,
            "nlc": False,
            "deriv": 1,
        },
        "scf": {
            "ansatz": 0,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "hf": {
            "ansatz": 0,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "hf+d": {
            "ansatz": 0,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "hf-3c": {
            "ansatz": 0,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "hf3c": {
            "ansatz": 0,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "pbeh-3c": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "pbeh3c": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "sogga11x": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "sogga11-x": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "mn12sx": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": True,
            "nlc": False,
            "deriv": 1,
        },
        "mn12-sx": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": True,
            "nlc": False,
            "deriv": 1,
        },
        "mn15": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "n12sx": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": True,
            "nlc": False,
            "deriv": 1,
        },
        "n12-sx": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": True,
            "nlc": False,
            "deriv": 1,
        },
        "pbe50": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "revpbe0": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "mpw1lyp": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "mpw1pbe": {
            "ansatz": 1,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "mgga-ms2h": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "ms2h": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "mgga_ms2h": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "mgga-mvsh": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "mvsh": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "mgga_mvsh": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "dldf": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "dldf+d": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "m05": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "m052x": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "m05-2x": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "m06": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "m062x": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "m06-2x": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "m06hf": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "m06-hf": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "m08hx": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "m08-hx": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "m08so": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "m08-so": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "m11": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": True,
            "nlc": False,
            "deriv": 1,
        },
        "scan0": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "revscan0": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "bmk": {
            "ansatz": 2,
            "c_hybrid": False,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "mp2mp2": {
            "ansatz": 0,
            "c_hybrid": True,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "mp2d": {
            "ansatz": 0,
            "c_hybrid": True,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "b2plyp": {
            "ansatz": 1,
            "c_hybrid": True,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "dsd-blyp": {
            "ansatz": 1,
            "c_hybrid": True,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "core-dsd-blyp": {
            "ansatz": 1,
            "c_hybrid": True,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "pbe02": {
            "ansatz": 1,
            "c_hybrid": True,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "pbe0-2": {
            "ansatz": 1,
            "c_hybrid": True,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "pbe0dh": {
            "ansatz": 1,
            "c_hybrid": True,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "pbe0-dh": {
            "ansatz": 1,
            "c_hybrid": True,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "dsdpbep86": {
            "ansatz": 1,
            "c_hybrid": True,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "dsd-pbep86": {
            "ansatz": 1,
            "c_hybrid": True,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "dsdpbepbe": {
            "ansatz": 1,
            "c_hybrid": True,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "dsd-pbepbe": {
            "ansatz": 1,
            "c_hybrid": True,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "b2gpplyp": {
            "ansatz": 1,
            "c_hybrid": True,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "pwpb95": {
            "ansatz": 2,
            "c_hybrid": True,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "ptpss": {
            "ansatz": 2,
            "c_hybrid": True,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
        "dsd-pbeb95": {
            "ansatz": 2,
            "c_hybrid": True,
            "x_hybrid": True,
            "c_lrc": False,
            "x_lrc": False,
            "nlc": False,
            "deriv": 1,
        },
    },
    "empirical_dispersion_suffixes": [
        "-chg",
        "-d2",
        "-d3mbj",
        "+d09",
        "-d3zero",
        "+d10",
        "-atmgr",
        "-d10",
        "-das2010",
        "-d09",
        "-d3mzero",
        "-d",
        "-dmp2",
        "-d3m(bj)",
        "-d3(bj)",
        "-nl",
        "-d3bj",
        "-d3m",
        "-d3",
        "-das2009",
    ],
}



import copy
from typing import Dict

import numpy as np
import pytest

import qcelemental
from qcelemental.testing import compare_molrecs

_string_prov_stamp = {"creator": "QCElemental", "version": "1.0", "routine": "qcelemental.molparse.from_string"}
_schema_prov_stamp = {"creator": "QCElemental", "version": "1.0", "routine": "qcelemental.molparse.from_schema"}

subject14 = """0 3\n--\nHe 0 0 -5\n--\n@He 0 0 5\nunits au"""

schema14_1 = {
    "schema_name": "qc_schema_input",
    "schema_version": 1,
    "molecule": {
        "validated": True,
        "geometry": [0.0, 0.0, -5.0, 0.0, 0.0, 5.0],
        "symbols": ["He", "He"],
        "atomic_numbers": [2, 2],
        "mass_numbers": [4, 4],
        "atom_labels": ["", ""],
        "fragments": [[0], [1]],
        "fragment_charges": [0.0, 0.0],
        "fragment_multiplicities": [3, 1],
        "masses": [4.00260325413, 4.00260325413],
        "name": "He2",
        "fix_com": False,
        "fix_orientation": False,
        "molecular_charge": 0.0,
        "molecular_multiplicity": 3,
        "real": [True, False],
    },
}

schema14_2 = {"schema_name": "qcschema_molecule", "schema_version": 2}
schema14_2.update(schema14_1["molecule"])

schema14_psi4 = {
    "geom": [0.0, 0.0, -5.0, 0.0, 0.0, 5.0],
    "elem": ["He", "He"],
    "elea": [4, 4],
    "elez": [2, 2],
    "fragment_charges": [0.0, 0.0],
    "fragment_multiplicities": [3, 1],
    "mass": [4.00260325413, 4.00260325413],
    "name": "He2",
    "fix_com": False,
    "fix_orientation": False,
    "molecular_charge": 0.0,
    "molecular_multiplicity": 3,
    "units": "Bohr",
    "fragment_separators": [1],
    "elbl": ["", ""],
    "real": [True, False],
}


def test_1_14a():
    fullans = copy.deepcopy(schema14_1)
    fullans["molecule"]["provenance"] = _string_prov_stamp

    final = qcelemental.molparse.from_string(subject14)
    kmol = qcelemental.molparse.to_schema(final["qm"], dtype=1)
    assert compare_molrecs(fullans["molecule"], kmol["molecule"])

    fullans = copy.deepcopy(schema14_psi4)
    fullans["provenance"] = _schema_prov_stamp

    molrec = qcelemental.molparse.from_schema(kmol)
    molrec = qcelemental.util.unnp(molrec)
    assert compare_molrecs(fullans, molrec)


def test_2_14b():
    fullans = copy.deepcopy(schema14_2)
    fullans["provenance"] = _string_prov_stamp

    final = qcelemental.molparse.from_string(subject14)
    kmol = qcelemental.molparse.to_schema(final["qm"], dtype=2)
    assert compare_molrecs(fullans, kmol)

    fullans = copy.deepcopy(schema14_psi4)
    fullans["provenance"] = _schema_prov_stamp

    molrec = qcelemental.molparse.from_schema(kmol)
    molrec = qcelemental.util.unnp(molrec)
    assert compare_molrecs(fullans, molrec)


def test_psi4_14c():
    fullans = copy.deepcopy(schema14_psi4)
    fullans["provenance"] = _string_prov_stamp

    final = qcelemental.molparse.from_string(subject14)
    kmol = qcelemental.molparse.to_schema(final["qm"], dtype="psi4")
    assert compare_molrecs(fullans, kmol)


def test_dtype_error():

    final = qcelemental.molparse.from_string(subject14)
    with pytest.raises(qcelemental.ValidationError) as e:
        qcelemental.molparse.to_schema(final["qm"], dtype="xkcd927")

    assert "Schema dtype not understood" in str(e.value)


@pytest.mark.parametrize("dtype", [1, 2])
def test_qcschema_ang_error(dtype):

    final = qcelemental.molparse.from_string(subject14)
    with pytest.raises(qcelemental.ValidationError) as e:
        qcelemental.molparse.to_schema(final["qm"], dtype=dtype, units="Angstrom")

    assert f"QCSchema {dtype} allows only 'Bohr' coordinates" in str(e.value)


def test_psi4_nm_error():

    final = qcelemental.molparse.from_string(subject14)
    with pytest.raises(qcelemental.ValidationError) as e:
        qcelemental.molparse.to_schema(final["qm"], dtype="psi4", units="nm")

    assert "Psi4 Schema psi4 allows only 'Bohr'/'Angstrom' coordinates, not nm" in str(e.value)


twobohrinang = 2.0 * qcelemental.constants.conversion_factor("bohr", "angstrom")
subject15 = """symmetry cS\nH 0 0 {twobohrinang}\nO 0 0 0\n2H_deut {twobohrinang} 0 0\nno_com\nno_reorient""".format(
    twobohrinang=twobohrinang
)

schema15_1: Dict = {
    "schema_name": "qc_schema_input",
    "schema_version": 1,
    "molecule": {
        "validated": True,
        "geometry": [0.0, 0.0, 2.0, 0.0, 0.0, 0.0, 2.0, 0.0, 0.0],
        "symbols": ["H", "O", "H"],
        "atomic_numbers": [1, 8, 1],
        "mass_numbers": [1, 16, 2],
        "atom_labels": ["", "", "_deut"],
        "fragments": [[0, 1, 2]],
        "fragment_charges": [0.0],
        "fragment_multiplicities": [1],
        "masses": [1.00782503223, 15.99491461957, 2.01410177812],
        "name": "H2O",
        "comment": "I has a comment",
        "fix_com": True,
        "fix_orientation": True,
        "fix_symmetry": "cs",
        "molecular_charge": 0.0,
        "molecular_multiplicity": 1,
        "real": [True, True, True],
    },
}

schema15_2 = {"schema_name": "qcschema_molecule", "schema_version": 2}
schema15_2.update(schema15_1["molecule"])

schema15_psi4 = {
    "geom": [0.0, 0.0, twobohrinang, 0.0, 0.0, 0.0, twobohrinang, 0.0, 0.0],
    "elem": ["H", "O", "H"],
    "elez": [1, 8, 1],
    "elea": [1, 16, 2],
    "elbl": ["", "", "_deut"],
    "fragment_charges": [0.0],
    "fragment_multiplicities": [1],
    "mass": [1.00782503223, 15.99491461957, 2.01410177812],
    "name": "H2O",
    "comment": "I has a comment",
    "fix_com": True,
    "fix_orientation": True,
    "fix_symmetry": "cs",
    "molecular_charge": 0.0,
    "molecular_multiplicity": 1,
    "units": "Angstrom",
    "fragment_separators": [],
    "real": [True, True, True],
}


def test_1_15a():
    fullans = copy.deepcopy(schema15_1)
    fullans["molecule"]["provenance"] = _string_prov_stamp

    final = qcelemental.molparse.from_string(subject15)
    final["qm"]["comment"] = "I has a comment"
    kmol = qcelemental.molparse.to_schema(final["qm"], dtype=1)
    assert compare_molrecs(fullans["molecule"], kmol["molecule"])

    fullans = copy.deepcopy(schema15_psi4)
    fullans["units"] = "Bohr"
    fullans["geom"] = [0.0, 0.0, 2.0, 0.0, 0.0, 0.0, 2.0, 0.0, 0.0]
    fullans["provenance"] = _schema_prov_stamp

    molrec = qcelemental.molparse.from_schema(kmol)
    molrec = qcelemental.util.unnp(molrec)
    assert compare_molrecs(fullans, molrec)


def test_2_15b():
    fullans = copy.deepcopy(schema15_2)
    fullans["provenance"] = _string_prov_stamp

    final = qcelemental.molparse.from_string(subject15)
    final["qm"]["comment"] = "I has a comment"
    kmol = qcelemental.molparse.to_schema(final["qm"], dtype=2)
    assert compare_molrecs(fullans, kmol)

    fullans = copy.deepcopy(schema15_psi4)
    fullans["units"] = "Bohr"
    fullans["geom"] = [0.0, 0.0, 2.0, 0.0, 0.0, 0.0, 2.0, 0.0, 0.0]
    fullans["provenance"] = _schema_prov_stamp

    molrec = qcelemental.molparse.from_schema(kmol)
    molrec = qcelemental.util.unnp(molrec)
    assert compare_molrecs(fullans, molrec)


def test_psi4_15c():
    fullans = copy.deepcopy(schema15_psi4)
    fullans["provenance"] = _string_prov_stamp

    final = qcelemental.molparse.from_string(subject15)
    final["qm"]["comment"] = "I has a comment"
    kmol = qcelemental.molparse.to_schema(final["qm"], dtype="psi4", units="Angstrom")
    assert compare_molrecs(fullans, kmol)


schema16_1 = {
    "schema_name": "qc_schema_input",
    "schema_version": 1,
    "molecule": {
        "validated": True,
        "geometry": [2.0, 2.0, 3.0],
        "symbols": ["C"],
        "masses": [12.0],
        "atom_labels": [""],
        "atomic_numbers": [6],
        "mass_numbers": [12],
        "real": [True],
        "name": "C",
        "molecular_charge": 0.0,
        "molecular_multiplicity": 1,
        "fragments": [[0]],
        "fragment_charges": [0.0],
        "fragment_multiplicities": [1],
        "fix_com": False,
        "fix_orientation": False,
        "provenance": {"creator": "Mystery Program", "version": "2018.3", "routine": "molecule builder"},
        "connectivity": [(0, 0, 0.0)],
    },
}

schema16_2 = {"schema_name": "qcschema_molecule", "schema_version": 2}
schema16_2.update(schema16_1["molecule"])

schema16_psi4 = {
    "units": "Bohr",
    "geom": np.array([2.0, 2.0, 3.0]),
    "elem": np.array(["C"]),
    "mass": np.array([12.0]),
    "elbl": np.array([""]),
    "elez": np.array([6]),
    "elea": np.array([12]),
    "real": np.array([True]),
    "name": "C",
    "molecular_charge": 0.0,
    "molecular_multiplicity": 1,
    "fragment_separators": [],
    "fragment_charges": [0.0],
    "fragment_multiplicities": [1],
    "fix_com": False,
    "fix_orientation": False,
    "provenance": {"creator": "Mystery Program", "version": "2018.3", "routine": "molecule builder"},
    "connectivity": [(0, 0, 0.0)],
}


def test_froto_1_16a():
    basic = {
        "schema_name": "qc_schema_output",
        "schema_version": 1,
        "molecule": {
            "geometry": [2, 2, 3],
            "symbols": ["C"],
            "connectivity": [(0.0, -0.0, 0)],
            "provenance": {"creator": "Mystery Program", "version": "2018.3", "routine": "molecule builder"},
        },
    }

    fullans = copy.deepcopy(schema16_1)
    fullans["molecule"]["provenance"] = _schema_prov_stamp

    roundtrip = qcelemental.molparse.to_schema(qcelemental.molparse.from_schema(basic), dtype=1)
    assert compare_molrecs(fullans["molecule"], roundtrip["molecule"])


def test_froto_2_16a():
    basic = {
        "schema_name": "qcschema_molecule",
        "schema_version": 2,
        "geometry": [2, 2, 3],
        "symbols": ["C"],
        "connectivity": [(0.0, -0.0, 0)],
        "provenance": {"creator": "Mystery Program", "version": "2018.3", "routine": "molecule builder"},
    }

    fullans = copy.deepcopy(schema16_2)
    fullans["provenance"] = _schema_prov_stamp

    roundtrip = qcelemental.molparse.to_schema(qcelemental.molparse.from_schema(basic), dtype=2)
    assert compare_molrecs(fullans, roundtrip)


@pytest.mark.parametrize("dtype", [1, 2])
def test_tofro_16b(dtype):

    fullans = copy.deepcopy(schema16_psi4)
    fullans["provenance"] = _schema_prov_stamp

    roundtrip = qcelemental.molparse.from_schema(qcelemental.molparse.to_schema(schema16_psi4, dtype=dtype))
    assert compare_molrecs(fullans, roundtrip)



import copy

import numpy as np
import pytest

import qcelemental as qcel
from qcelemental.testing import compare_molrecs

_schema_prov_stamp = {"creator": "QCElemental", "version": "1.0", "routine": "qcelemental.molparse.from_schema"}


@pytest.mark.parametrize(
    "inp,expected",
    [
        (
            {"frag_pattern": [[0], [1]], "geom": [0.0, 0.0, 0.0, 1.0, 0.0, 0.0], "elbl": ["O", "H"]},
            {
                "fragment_separators": np.array([1]),
                "geom": np.array([0.0, 0.0, 0.0, 1.0, 0.0, 0.0]),
                "elbl": np.array(["O", "H"]),
            },
        ),
        (
            {
                "frag_pattern": [[2, 0], [1]],
                "geom": np.array([[0.0, 0.0, 1.0], [0.0, 0.0, 2.0], [0.0, 0.0, 0.0]]),
                "elem": np.array(["Li", "H", "He"]),
            },
            {
                "fragment_separators": np.array([2]),
                "geom": np.array([0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 2.0]),
                "elem": np.array(["He", "Li", "H"]),
            },
        ),
        (
            {"frag_pattern": [[2, 0], [1]], "elez": [3, 1, 2]},
            {"fragment_separators": np.array([2]), "elez": np.array([2, 3, 1])},
        ),
    ],
)
def test_contiguize_from_fragment_pattern(inp, expected):
    ans = qcel.molparse.contiguize_from_fragment_pattern(**inp)

    # compare_molrecs instead of compare_dicts handles some fragment_separators types issues
    assert compare_molrecs(expected, ans, atol=1.0e-6)


@pytest.mark.parametrize(
    "inp,expected",
    [
        (
            {
                "frag_pattern": [[2, 0], [1, 3]],
                "geom": np.array([[0.0, 0.0, 1.0], [0.0, 0.0, 2.0], [0.0, 0.0, 0.0]]),
                "elem": np.array(["Li", "H", "He"]),
            },
            "dropped atoms",
        ),
        ({"frag_pattern": [[2, 0], [1, 4]]}, "Fragmentation pattern skips atoms"),
        (
            {
                "frag_pattern": [[2, 0], [1, 3]],
                "elem": np.array(["U", "Li", "H", "He"]),
                "elbl": np.array(["Li", "H", "He"]),
            },
            "wrong number of atoms in array",
        ),
        (
            {"frag_pattern": [[2, 0], [1]], "elez": [3, 1, 2], "throw_reorder": True},
            "reorder atoms to accommodate non-contiguous fragments",
        ),
    ],
)
def test_contiguize_from_fragment_pattern_error(inp, expected):
    with pytest.raises(qcel.ValidationError) as e:
        qcel.molparse.contiguize_from_fragment_pattern(**inp)

    assert expected in str(e.value)


schema14_1 = {
    "geometry": [0.0, 0.0, -5.0, 0.0, 0.0, 5.0],
    "symbols": ["He", "He"],
    "fragments": [[0], [1]],
    "fragment_charges": [0.0, 0.0],
    "fragment_multiplicities": [3, 1],
    "masses": [4.00260325413, 4.00260325413],
    "name": "He2",
    "fix_com": False,
    "fix_orientation": False,
    "molecular_charge": 0.0,
    "molecular_multiplicity": 3,
    "real": [True, False],
}

schema14_psi4_np = {
    "geom": np.array([0.0, 0.0, -5.0, 0.0, 0.0, 5.0]),
    "elem": np.array(["He", "He"]),
    "elea": np.array([4, 4]),
    "elez": np.array([2, 2]),
    "fragment_charges": [0.0, 0.0],
    "fragment_multiplicities": [3, 1],
    "mass": np.array([4.00260325413, 4.00260325413]),
    "name": "He2",
    "fix_com": False,
    "fix_orientation": False,
    "molecular_charge": 0.0,
    "molecular_multiplicity": 3,
    "units": "Bohr",
    "fragment_separators": [1],
    "elbl": np.array(["", ""]),
    "real": np.array([True, False]),
    "provenance": _schema_prov_stamp,
}


def test_from_schema_1_14e():
    schema = {"schema_name": "qc_schema", "schema_version": 1, "molecule": copy.deepcopy(schema14_1)}

    ans = qcel.molparse.from_schema(schema)
    assert compare_molrecs(schema14_psi4_np, ans, 4)


def test_from_schema_1p5_14e():
    # this is the oddball case where passing code has internally a dtype=2 molecule
    #   but it's still passing the outer data structure
    schema = {"schema_name": "qc_schema", "schema_version": 1, "molecule": copy.deepcopy(schema14_1)}
    schema["molecule"].update({"schema_name": "qcschema_molecule", "schema_version": 2})

    ans = qcel.molparse.from_schema(schema)
    assert compare_molrecs(schema14_psi4_np, ans, 4)


def test_from_schema_2_14e():
    schema = copy.deepcopy(schema14_1)
    schema.update({"schema_name": "qcschema_molecule", "schema_version": 2})

    ans = qcel.molparse.from_schema(schema)
    assert compare_molrecs(schema14_psi4_np, ans, 4)


def test_from_schema_error_f():
    schema = {"schema_name": "private_schema", "schema_version": 1, "molecule": copy.deepcopy(schema14_1)}

    with pytest.raises(qcel.ValidationError) as e:
        qcel.molparse.from_schema(schema)

    assert "Schema not recognized" in str(e.value)


def test_from_schema_1_nfr_error_14g():
    schema = {"schema_name": "qc_schema", "schema_version": 1, "molecule": copy.deepcopy(schema14_1)}
    schema["molecule"].pop("fragments")

    with pytest.raises(qcel.ValidationError) as e:
        qcel.molparse.from_schema(schema)

    assert "Dimension mismatch among fragment quantities: sep + 1 (1), chg (2), and mult(2)" in str(e.value)



import pytest

import qcelemental
from qcelemental.testing import compare, compare_values


@pytest.mark.parametrize(
    "inp,expected",
    [
        ("@ca_miNe", {"E": "ca", "Z": None, "user": "_miNe", "A": None, "real": False, "mass": None}),
        ("Gh(Ca_mine)", {"E": "Ca", "Z": None, "user": "_mine", "A": None, "real": False, "mass": None}),
        ("@Ca_mine@1.07", {"E": "Ca", "Z": None, "user": "_mine", "A": None, "real": False, "mass": 1.07}),
        ("Gh(cA_MINE@1.07)", {"E": "cA", "Z": None, "user": "_MINE", "A": None, "real": False, "mass": 1.07}),
        ("@40Ca_mine@1.07", {"E": "Ca", "Z": None, "user": "_mine", "A": 40, "real": False, "mass": 1.07}),
        ("Gh(40Ca_mine@1.07)", {"E": "Ca", "Z": None, "user": "_mine", "A": 40, "real": False, "mass": 1.07}),
        ("444lu333@4.0", {"E": "lu", "Z": None, "user": "333", "A": 444, "real": True, "mass": 4.0}),
        ("@444lu333@4.4", {"E": "lu", "Z": None, "user": "333", "A": 444, "real": False, "mass": 4.4}),
        ("8i", {"E": "i", "Z": None, "user": None, "A": 8, "real": True, "mass": None}),
        ("53_mI4", {"Z": 53, "E": None, "user": "_mI4", "A": None, "real": True, "mass": None}),
        ("@5_MINEs3@4.4", {"Z": 5, "E": None, "user": "_MINEs3", "A": None, "real": False, "mass": 4.4}),
        ("Gh(555_mines3@0.1)", {"Z": 555, "E": None, "user": "_mines3", "A": None, "real": False, "mass": 0.1}),
    ],
)
def test_parse_nucleus_label(inp, expected):
    lbl_A, lbl_Z, lbl_E, lbl_mass, lbl_real, lbl_user = qcelemental.molparse.nucleus.parse_nucleus_label(inp)

    assert compare(expected["real"], lbl_real, inp + " real")
    assert compare(expected["A"], lbl_A, inp + " A")
    assert compare(expected["Z"], lbl_Z, inp + " Z")
    assert compare(expected["E"], lbl_E, inp + " symbol")
    assert compare(expected["user"], lbl_user, inp + " user")
    assert compare_values(expected["mass"], lbl_mass, inp + " mass", passnone=True, atol=1.0e-6)


@pytest.mark.parametrize("inp", ["1L1A1B1"])
def test_parse_nucleus_label_error(inp):
    with pytest.raises(qcelemental.ValidationError):
        qcelemental.molparse.nucleus.parse_nucleus_label(inp)



import json
from pathlib import Path
from typing import Any, Dict, Optional, Set, Union

import numpy as np
from pydantic import BaseModel, BaseSettings

from qcelemental.testing import compare_recursive
from qcelemental.util import deserialize, serialize
from qcelemental.util.autodocs import AutoPydanticDocGenerator


def _repr(self) -> str:
    return f'{self.__repr_name__()}({self.__repr_str__(", ")})'


class ProtoModel(BaseModel):
    class Config:
        allow_mutation: bool = False
        extra: str = "forbid"
        json_encoders: Dict[str, Any] = {np.ndarray: lambda v: v.flatten().tolist()}
        serialize_default_excludes: Set = set()
        serialize_skip_defaults: bool = False
        force_skip_defaults: bool = False

    def __init_subclass__(cls, **kwargs) -> None:
        super().__init_subclass__(**kwargs)
        cls.__doc__ = AutoPydanticDocGenerator(cls, always_apply=True)

        if "pydantic" in cls.__repr__.__module__:
            cls.__repr__ = _repr

        if "pydantic" in cls.__str__.__module__:
            cls.__str__ = _repr

    @classmethod
    def parse_raw(cls, data: Union[bytes, str], *, encoding: str = None) -> "ProtoModel":  # type: ignore
        """
        Parses raw string or bytes into a Model object.

        Parameters
        ----------
        data : Union[bytes, str]
            A serialized data blob to be deserialized into a Model.
        encoding : str, optional
            The type of the serialized array, available types are: {'json', 'json-ext', 'msgpack-ext', 'pickle'}

        Returns
        -------
        Model
            The requested model from a serialized format.
        """

        if encoding is None:
            if isinstance(data, str):
                encoding = "json"
            elif isinstance(data, bytes):
                encoding = "msgpack-ext"
            else:
                raise TypeError("Input is neither str nor bytes, please specify an encoding.")

        if encoding.endswith(("json", "javascript", "pickle")):
            return super().parse_raw(data, content_type=encoding)
        elif encoding in ["msgpack-ext", "json-ext"]:
            obj = deserialize(data, encoding)
        else:
            raise TypeError(f"Content type '{encoding}' not understood.")

        return cls.parse_obj(obj)

    @classmethod
    def parse_file(cls, path: Union[str, Path], *, encoding: str = None) -> "ProtoModel":  # type: ignore
        """Parses a file into a Model object.

        Parameters
        ----------
        path : Union[str, Path]
            The path to the file.
        encoding : str, optional
            The type of the files, available types are: {'json', 'msgpack', 'pickle'}. Attempts to
            automatically infer the file type from the file extension if None.

        Returns
        -------
        Model
            The requested model from a serialized format.

        """
        path = Path(path)
        if encoding is None:
            if path.suffix in [".json", ".js"]:
                encoding = "json"
            elif path.suffix in [".msgpack"]:
                encoding = "msgpack-ext"
            elif path.suffix in [".pickle"]:
                encoding = "pickle"
            else:
                raise TypeError("Could not infer `encoding`, please provide a `encoding` for this file.")

        return cls.parse_raw(path.read_bytes(), encoding=encoding)

    def dict(self, **kwargs) -> Dict[str, Any]:
        encoding = kwargs.pop("encoding", None)

        kwargs["exclude"] = (
            kwargs.get("exclude", None) or set()
        ) | self.__config__.serialize_default_excludes  # type: ignore
        kwargs.setdefault("exclude_unset", self.__config__.serialize_skip_defaults)  # type: ignore
        if self.__config__.force_skip_defaults:  # type: ignore
            kwargs["exclude_unset"] = True

        data = super().dict(**kwargs)

        if encoding is None:
            return data
        elif encoding == "json":
            return json.loads(serialize(data, encoding="json"))
        else:
            raise KeyError(f"Unknown encoding type '{encoding}', valid encoding types: 'json'.")

    def serialize(
        self,
        encoding: str,
        *,
        include: Optional[Set[str]] = None,
        exclude: Optional[Set[str]] = None,
        exclude_unset: Optional[bool] = None,
    ) -> Union[bytes, str]:
        """Generates a serialized representation of the model

        Parameters
        ----------
        encoding : str
            The serialization type, available types are: {'json', 'json-ext', 'msgpack-ext'}
        include : Optional[Set[str]], optional
            Fields to be included in the serialization.
        exclude : Optional[Set[str]], optional
            Fields to be excluded in the serialization.
        exclude_unset : Optional[bool], optional
            If True, skips fields that have default values provided.

        Returns
        -------
        Union[bytes, str]
            The serialized model.
        """

        kwargs = {}
        if include:
            kwargs["include"] = include
        if exclude:
            kwargs["exclude"] = exclude
        if exclude_unset:
            kwargs["exclude_unset"] = exclude_unset

        data = self.dict(**kwargs)

        return serialize(data, encoding=encoding)

    def json(self, **kwargs):
        # Alias JSON here from BaseModel to reflect dict changes
        return self.serialize("json", **kwargs)

    def compare(self, other: Union["ProtoModel", BaseModel], **kwargs) -> bool:
        """Compares the current object to the provided object recursively.

        Parameters
        ----------
        other : Model
            The model to compare to.
        **kwargs
            Additional kwargs to pass to ``qcelemental.compare_recursive``.

        Returns
        -------
        bool
            True if the objects match.
        """
        return compare_recursive(self, other, **kwargs)


class AutodocBaseSettings(BaseSettings):
    def __init_subclass__(cls) -> None:
        cls.__doc__ = AutoPydanticDocGenerator(cls, always_apply=True)



import collections
from typing import Any, Dict, List, Tuple, Union

import numpy as np

from ..molutil import guess_connectivity
from ..physical_constants import constants


def to_string(
    molrec: Dict,
    dtype: str,
    units: str = None,
    *,
    atom_format: str = None,
    ghost_format: str = None,
    width: int = 17,
    prec: int = 12,
    return_data: bool = False,
) -> Union[str, Tuple[str, Dict]]:
    """Format a string representation of QM molecule.

    Parameters
    ----------
    molrec : Dict
        Psi4 json Molecule spec.
    dtype : str, {'xyz', 'cfour', 'nwchem', 'molpro', 'orca', 'turbomole',
                  'qchem'}
        Overall string format. Note that it's possible to request variations
        that don't fit the dtype spec so may not be re-readable (e.g., ghost
        and mass in nucleus label with ``'xyz'``).
        'cfour' forces nucleus label, ignoring atom_format, ghost_format
    units : str, optional
        Units in which to write string. Usually ``Angstrom`` or ``Bohr``
        but may be any length unit.  There is not an option to write in
        intrinsic/input units. For ``dtype='xyz', units='Bohr'`` where the
        format doesn't have a slot to specify units, "au" is added so that
        readable as ``dtype='xyz+'``.
    atom_format : str, optional
        General format is ``'{elem}'``. A format string that may contain fields
        'elea' (-1 will be ''), 'elez', 'elem', 'mass', 'elbl' in any
        arrangement. For example, if a format naturally uses element symbol
        and you want atomic number instead with mass info, too, pass
        ``'{elez}@{mass}'``. See `ghost_format` for handling field 'real'.
    ghost_format : str, optional
        General format is ``'@{elem}'``. Like `atom_format`, but this formatter
        is used when `real=False`. To suppress ghost atoms, use `ghost_format=''`.
    width : int, optional
        Field width for formatting coordinate float.
    prec : int, optional
        Number of decimal places for formatting coordinate float.
    return_data : bool, optional
        Whether to return dictionary with additional info from the molrec that's
        not expressible in the string but may be of interest to the QC program.
        Note that field names are in QCSchema, not molrec, language.

    Returns
    -------
    str
        String representation of the molecule.
    str, dict
        When ``return_data=True`, return additionally a dictionary
            keywords: key, value pairs for processing molecule info into options
            fields: aspects of ``qcelemental.models.Molecule`` expressed into
                    string _or_ keywords.

    """

    # funits, fiutau = process_units(molrec)
    # molrec = self.to_dict(force_units=units, np_out=True)

    dtype = dtype.lower()

    default_units = {
        "xyz": "Angstrom",
        "xyz+": "Angstrom",
        "nglview-sdf": "Angstrom",
        "cfour": "Bohr",
        "gamess": "Bohr",
        "molpro": "Bohr",
        "nwchem": "Bohr",
        "orca": "Bohr",
        "psi4": "Bohr",
        "qchem": "Bohr",
        "terachem": "Bohr",
        "turbomole": "Bohr",
        "madness": "Bohr",  # madness by default reads au and optionally can read angs/angstrom
        "mrchem": "Bohr",  # MRChem by default reads au and optionally can read angs/angstrom
    }
    if dtype not in default_units:
        raise KeyError(f"dtype '{dtype}' not understood.")

    # Handle units
    if units is None:
        units = default_units[dtype]

    if molrec["units"] == "Angstrom" and units.capitalize() == "Angstrom":
        factor = 1.0
    elif molrec["units"] == "Angstrom" and units.capitalize() == "Bohr":
        if "input_units_to_au" in molrec:
            factor = molrec["input_units_to_au"]
        else:
            factor = 1.0 / constants.bohr2angstroms
    elif molrec["units"] == "Bohr" and units.capitalize() == "Angstrom":
        factor = constants.bohr2angstroms
    elif molrec["units"] == "Bohr" and units.capitalize() == "Bohr":
        factor = 1.0
    else:
        factor = constants.conversion_factor(molrec["units"], units)
    geom = np.asarray(molrec["geom"]).reshape((-1, 3)) * factor

    name = molrec.get("name", formula_generator(molrec["elem"]))
    tagline = """auto-generated by QCElemental from molecule {}""".format(name)

    class Data:
        fields: List[str] = ["atomic_numbers", "geometry", "symbols"]
        keywords: Dict[str, Any] = {}

        _dict_attrs: List[str] = ["fields", "keywords"]

        def to_dict(self) -> Dict:
            return {attr: getattr(self, attr) for attr in self._dict_attrs}

    data = Data()

    if dtype in ["xyz", "xyz+"]:
        # Notes
        # * if units not in umap (e.g., nm), can't be read back in by from_string()

        atom_format = "{elem}" if atom_format is None else atom_format
        ghost_format = "@{elem}" if ghost_format is None else ghost_format
        umap = {"bohr": "au", "angstrom": ""}

        atoms = _atoms_formatter(molrec, geom, atom_format, ghost_format, width, prec, 2)
        nat = len(atoms)

        first_line = """{} {}""".format(str(nat), umap.get(units.lower(), units.lower()))
        smol = [first_line.rstrip()]
        smol.append(f"{int(molrec['molecular_charge'])} {molrec['molecular_multiplicity']} {name}")

        smol.extend(atoms)

    elif dtype == "orca":

        atom_format = "{elem}"
        ghost_format = "{elem}:"
        umap = {"bohr": "! Bohrs", "angstrom": "!"}

        atoms = _atoms_formatter(molrec, geom, atom_format, ghost_format, width, prec, 2)

        smol = []
        smol.append(umap[units.lower()])
        smol.append("")
        smol.append(f"*xyz {int(molrec['molecular_charge'])} {molrec['molecular_multiplicity']}")
        smol.extend(atoms)
        smol.append("*")

    elif dtype == "cfour":
        # Notes
        # * losing identity of ghost atoms. picked up again in basis formatting
        # * casting 'molecular_charge' to int
        # * no spaces at the beginning of 1st/comment line is important

        atom_format = "{elem}"
        ghost_format = "GH"
        # TODO handle which units valid
        umap = {"bohr": "bohr", "angstrom": "angstrom"}

        atoms = _atoms_formatter(molrec, geom, atom_format, ghost_format, width, prec, 2)

        smol = [tagline]
        smol.extend(atoms)

        data.fields.extend(["molecular_charge", "molecular_multiplicity"])
        data.keywords = {
            "charge": int(molrec["molecular_charge"]),
            "multiplicity": molrec["molecular_multiplicity"],
            "units": umap.get(units.lower()),
            "coordinates": "cartesian",
        }

    elif dtype == "molpro":

        atom_format = "{elem}"
        ghost_format = "{elem}"
        umap = {"bohr": "bohr", "angstrom": "angstrom"}

        atoms = _atoms_formatter(molrec, geom, atom_format, ghost_format, width, prec, 2)

        smol = []

        # Don't orient the molecule if asked to fix_com or fix_orientation
        if molrec["fix_orientation"] or molrec["fix_com"]:
            smol.append("{orient,noorient}")

        # Have no symmetry if asked to fix_symmetry
        if "fix_symmetry" in molrec.keys() and molrec["fix_symmetry"] == "c1":
            smol.append("{symmetry,nosym}")
        elif "fix_symmetry" not in molrec.keys():
            smol.append("{symmetry,auto}")

        smol.append("")

        units_line = f"""{{{umap.get(units.lower())}}}"""
        geom_line = """geometry={"""
        end_bracket = """}"""
        smol.append(units_line)
        smol.append(geom_line)
        smol.extend(atoms)
        smol.append(end_bracket)

        # Write ghost atom declarations in Molpro (using dummy card)
        if False in molrec["real"]:
            ghost_line = "dummy," + ",".join([str(idx + 1) for idx, real in enumerate(molrec["real"]) if not real])
            smol.append(ghost_line)

        smol.append(f"set,charge={molrec['molecular_charge']}")
        # The Molpro "spin" is the multiplicity minus one
        smol.append(f"set,spin={molrec['molecular_multiplicity']-1}")

    elif dtype == "nwchem":

        atom_format = "{elem}"
        ghost_format = "GH"
        # TODO handle which units valid
        umap = {"bohr": "bohr", "angstrom": "angstroms", "nm": "nanometers", "pm": "picometers"}

        atoms = _atoms_formatter(molrec, geom, atom_format, ghost_format, width, prec, 2)

        first_line = f"""geometry units {umap.get(units.lower())}"""
        # noautosym nocenter  # no reorienting input geometry
        fix_symm = molrec.get("fix_symmetry", None)
        symm_line = ""
        if fix_symm:
            symm_line = "symmetry {}".format(fix_symm)  # not quite what Jiyoung had
        last_line = """end"""
        smol = [first_line]
        smol.extend(atoms)
        smol.append(symm_line)
        smol.append(last_line)

        data.fields.extend(["molecular_charge", "molecular_multiplicity"])
        data.keywords = {"charge": int(molrec["molecular_charge"])}
        if molrec["molecular_multiplicity"] != 1:
            data.keywords["scf__nopen"] = molrec["molecular_multiplicity"] - 1
            data.keywords["dft__mult"] = molrec["molecular_multiplicity"]

    elif dtype == "madness":

        atom_format = "{elem}"
        ghost_format = "GH"
        # TODO handle which units valid
        umap = {"bohr": "au", "angstrom": "angstrom"}

        atoms = _atoms_formatter(molrec, geom, atom_format, ghost_format, width, prec, 2)

        first_line = f"""geometry"""
        second_line = f"""units {umap.get(units.lower())}"""
        last_line = """end"""
        # noautosym nocenter  # no reorienting input geometry
        smol = [first_line]
        smol.append(second_line)
        smol.extend(atoms)
        smol.append(last_line)

        data.fields.extend(["molecular_charge", "molecular_multiplicity"])
        data.keywords = {"charge": int(molrec["molecular_charge"])}
        if molrec["molecular_multiplicity"] != 1:
            data.keywords["spin_restricted"] = "false"

    elif dtype == "gamess":

        atom_format = " {elem}{elbl} {elez}"
        ghost_format = " {BQ} -{elez}"
        umap = {"bohr": "bohr", "angstrom": "angs"}

        atoms = _atoms_formatter(molrec, geom, atom_format, ghost_format, width, prec, 2)

        first_line = """ $data"""
        second_line = f""" {tagline}"""
        third_line = """ C1"""
        last_line = """ $end"""

        smol = [first_line, second_line, third_line]
        smol.extend(atoms)
        smol.append(last_line)

        data.fields.extend(["molecular_charge", "molecular_multiplicity"])
        data.keywords = {
            "contrl__icharg": int(molrec["molecular_charge"]),
            "contrl__mult": molrec["molecular_multiplicity"],
            "contrl__units": umap.get(units.lower()),
            "contrl__coord": "prinaxis",
        }

    elif dtype == "terachem":

        atom_format = "{elem}"
        ghost_format = "X{elem}"
        umap = {"bohr": "au", "angstrom": ""}

        atoms = _atoms_formatter(molrec, geom, atom_format, ghost_format, width, prec, 2)

        first_line = f"""{len(atoms)} {umap[units.lower()]}"""
        smol = [first_line.rstrip(), name]
        smol.extend(atoms)

    elif dtype == "psi4":

        atom_format = "{elem}{elbl}"
        ghost_format = "Gh({elem}{elbl})"
        umap = {"bohr": "bohr", "angstrom": "angstrom"}

        atoms = _atoms_formatter(molrec, geom, atom_format, ghost_format, width, prec, 2)

        smol = [f"""{int(molrec['molecular_charge'])} {molrec['molecular_multiplicity']}"""]
        split_atoms = np.split(atoms, molrec["fragment_separators"])
        for ifr, fr in enumerate(split_atoms):
            if len(split_atoms) > 1:  # harmless to include but tidier to exclude
                smol.extend(["--", f"{int(molrec['fragment_charges'][ifr])} {molrec['fragment_multiplicities'][ifr]}"])
            smol.extend(fr.tolist())

        # append units and any other non-default molecule keywords
        smol.append(f"units {umap[units.lower()]}")
        if molrec["fix_com"]:
            smol.append("no_com")
        if molrec["fix_orientation"]:
            smol.append("no_reorient")

        data.fields.extend(
            [
                "molecular_charge",
                "molecular_multiplicity",
                "fragments",
                "fragment_charges",
                "fragment_multiplicities",
                "fix_com",
                "fix_orientation",
                "real",
            ]
        )
        data.keywords = {}

    elif dtype == "turbomole":
        # In Turbomole coord files the coordinates come first, and the atomic
        # symbol comes afterwards.
        # Handling of ghost atoms is done in the basis section of the control
        # file by setting the nuclear charge of certain atoms to zero.
        atom_format = "{elem}"
        ghost_format = "{elem}"
        umap = {"bohr": "bohr"}
        umap[units.lower()]  # trigger error if downstream can't handle

        atoms = _atoms_formatter(molrec, geom, atom_format, ghost_format, width, prec, 2, xyze=True)
        atoms = [at.lower() for at in atoms]

        smol = ["$coord"] + atoms + ["$end"]

    elif dtype == "nglview-sdf":
        # SDF is pretty special, handle it manually

        if units.capitalize() != "Angstrom":
            raise ValueError("SDF Format must be in Angstroms")

        ghost_format = ghost_format or "Gh"

        connectivity = molrec.get("connectivity", None)
        if connectivity is None:
            bohr_geom = geom * constants.conversion_factor("Angstrom", "Bohr")
            connectivity = guess_connectivity(molrec["elem"], bohr_geom, default_connectivity=1)

        smol = []
        smol.append("")
        smol.append("QCElemental\n")
        smol.append(f"{len(molrec['real']):3d} {len(connectivity):2d}  0  0  0  0  0  0  0  0  0")
        for real, sym, xyz in zip(molrec["real"], molrec["elem"], geom):
            if bool(real) is False:
                sym = ghost_format
            smol.append(f"{xyz[0]:10.4f}{xyz[1]:10.4f}{xyz[2]:10.4f}{sym:>3s}  0  0     0  0  0  0  0  0")

        for a1, a2, b in connectivity:
            smol.append(f" {(a1 + 1):2d} {(a2 + 1):2d}  {int(b):1d}  0  0  0  0")

    elif dtype == "qchem":

        atom_format = "{elem}"
        ghost_format = "@{elem}"
        umap = {"bohr": True, "angstrom": False}

        atoms = _atoms_formatter(molrec, geom, atom_format, ghost_format, width, prec, 2)

        first_line = "$molecule"
        chgmult_line = f"""{int(molrec['molecular_charge'])} {molrec['molecular_multiplicity']}"""
        last_line = "$end"

        smol = [first_line, chgmult_line]
        split_atoms = np.split(atoms, molrec["fragment_separators"])
        for ifr, fr in enumerate(split_atoms):
            if len(split_atoms) > 1:
                smol.extend(
                    ["--", f"""{int(molrec['fragment_charges'][ifr])} {molrec['fragment_multiplicities'][ifr]}"""]
                )
            smol.extend(fr.tolist())
        smol.append(last_line)

        data.fields.extend(
            [
                "fix_com",
                "fix_orientation",
                "fragment_charges",
                "fragment_multiplicities",
                "molecular_charge",
                "molecular_multiplicity",
                "real",
                "units",
            ]
        )

        data.keywords = {
            "no_reorient": molrec["fix_orientation"] or molrec["fix_com"],
            "input_bohr": umap[units.lower()],
        }

        if "fix_symmetry" in molrec.keys() and molrec["fix_symmetry"] == "c1":
            data.keywords["sym_ignore"] = True
            data.keywords["symmetry"] = False

    elif dtype == "mrchem":
        atom_format = "{elem}"
        ghost_format = "{elem}"
        atoms = _atoms_formatter(molrec, geom, atom_format, ghost_format, width, prec, 2)

        smol = (
            [
                "Molecule {",
                f"charge = {int(molrec['molecular_charge'])}",
                f"multiplicity = {molrec['molecular_multiplicity']}",
                f"translate = {molrec['fix_com']}",
                "$coords",
            ]
            + atoms
            + ["$end\n}"]
        )

        # this is what we are actually interested in:
        # when dtype="mrchem", we always want to set return_data=True
        data.keywords = {
            "charge": int(molrec["molecular_charge"]),
            "multiplicity": molrec["molecular_multiplicity"],
            "translate": molrec["fix_com"],
            "coords": "\n".join(atoms),
        }

    else:
        raise KeyError(f"dtype '{dtype}' not understood.")

    smol_ret = "\n".join(smol) + "\n"
    if return_data:
        return smol_ret, data.to_dict()
    else:
        return smol_ret


def _atoms_formatter(molrec, geom, atom_format, ghost_format, width, prec, sp, xyze=False):
    """Format a list of strings, one per atom from `molrec`."""

    nat = geom.shape[0]
    fxyz = """{:>{width}.{prec}f}"""
    sp = """{:{sp}}""".format("", sp=sp)

    atoms = []
    for iat in range(nat):
        atom = []
        atominfo = {
            "elea": "" if molrec["elea"][iat] == -1 else molrec["elea"][iat],
            "elez": molrec["elez"][iat],
            "elem": molrec["elem"][iat],
            "mass": molrec["mass"][iat],
            "elbl": molrec["elbl"][iat],
        }

        if molrec["real"][iat]:
            nuc = """{:{width}}""".format(atom_format.format(**atominfo), width=width)
            atom.append(nuc)
        else:
            if ghost_format in ["", None]:
                continue
            else:
                nuc = """{:{width}}""".format(ghost_format.format(**atominfo), width=width)
                atom.append(nuc)

        atom.extend([fxyz.format(x, width=width, prec=prec) for x in geom[iat]])
        if xyze:
            atom.append(atom.pop(0).rstrip())
        atoms.append(sp.join(atom))

    return atoms


def formula_generator(elem):
    """Return simple chemical formula from element list `elem`.

    >>> formula_generator(['C', 'Ca', 'O', 'O', 'Ag']
    AgCCaO2

    """
    counted = collections.Counter(elem)
    return "".join((el if cnt == 1 else (el + str(cnt))) for el, cnt in sorted(counted.items()))



from .autodocs import auto_gen_docs_on_demand, get_base_docs
from .gph_uno_bipartite import uno
from .importing import parse_version, safe_version, which, which_import
from .internal import provenance_stamp
from .itertools import unique_everseen

# from .mpl import plot_coord
from .misc import (
    compute_angle,
    compute_dihedral,
    compute_distance,
    distance_matrix,
    filter_comments,
    measure_coordinates,
    standardize_efp_angles_units,
    unnp,
    update_with_error,
)
from .np_blockwise import blockwise_contract, blockwise_expand
from .np_rand3drot import random_rotation_matrix
from .scipy_hungarian import linear_sum_assignment
from .serialization import (
    deserialize,
    json_dumps,
    json_loads,
    jsonext_dumps,
    jsonext_loads,
    msgpackext_dumps,
    msgpackext_loads,
    serialize,
)



from decimal import Decimal

import pytest

import qcelemental

_pc_default = qcelemental.constants.name[-4:]


@pytest.fixture(scope="module", params=[None, "CODATA2014", "CODATA2018"])
def contexts(request):
    if request.param:
        return qcelemental.PhysicalConstantsContext(request.param)
    else:
        return qcelemental.constants


@pytest.fixture(scope="module")
def constants2014(request):
    return qcelemental.PhysicalConstantsContext("CODATA2014")


@pytest.fixture(scope="module")
def constants2018(request):
    return qcelemental.PhysicalConstantsContext("CODATA2018")


def _assert_forward_conversion(expected, converted):
    # Build values and quantize to the tolerance
    expected = Decimal(expected)
    from_to_value = Decimal(converted).quantize(expected)
    assert from_to_value == expected


def _assert_inverse_conversion(expected, converted):
    # Build inverse
    expected = Decimal(expected)
    inv_expected = 1 / Decimal(expected)
    to_from_value = Decimal(converted)

    # Expected to a relative tolerance as the number of digits plus one for rounding
    # Using float comparisons as we are taking an (1 / float) inverse in the conversion code
    rel_tol = float("10e-{}".format(len(expected.as_tuple().digits)))
    assert pytest.approx(float(inv_expected), rel_tol) == float(to_from_value)


# 2014 CODATA values
@pytest.mark.parametrize(
    "from_unit, to_unit, expected",
    [
        # First check the 7 canonical NIST conversions
        ("hartree", "Hz", "6.579683920711e15"),
        ("amu", "hartree", "3.4231776902e7"),
        ("hartree", "kilogram", "4.850870129e-35"),
        ("hartree", "J/mol", "2625499.638"),
        ("hartree", "eV", "27.21138602"),
        ("hartree", "wavenumber", "2.194746313702e5"),
        ("hartree", "kelvin", "3.1577513e5"),
        # Check other values
        ("millihartree", "Hz", "6.579683920711e9"),
        ("mhartree", "Hz", "6.579683920711e9"),
        ("hartree", "microgram", "4.850870129e-26"),
        ("microhartree", "wavenumber", "2.194746313702e-7"),
        ("uhartree", "wavenumber", "2.194746313702e-7"),
        ("kilohartree", "kiloeV", "27.21138602"),
        ("hartree", "kiloeV", "0.02721138602"),
        ("millihartree", "J/mol", "2625.499638"),
        ("hartree", "kJ/mol", "2625.499638"),
        ("hartree", "J", "4.359744650e-18"),
        ("1/m", "hartree", "4.556335e-8"),
    ],
)
def test_unit_conversion_nist_2014(from_unit, to_unit, expected, constants2014):
    converted = constants2014.conversion_factor(from_unit, to_unit)
    _assert_forward_conversion(expected, converted)


# 2018 CODATA values
@pytest.mark.parametrize(
    "from_unit, to_unit, expected",
    [
        # First check the 7 canonical NIST conversions
        ("hartree", "Hz", "6.579683920502e15"),
        ("amu", "hartree", "3.4231776874e7"),  # potentially another sigfig in 2018
        ("hartree", "kilogram", "4.8508702095432e-35"),
        ("hartree", "J/mol", "2625499.6394798"),
        ("hartree", "eV", "27.211386245988"),
        ("hartree", "wavenumber", "2.1947463136320e5"),
        ("hartree", "kelvin", "3.1577502480407e5"),
        # Check other values
        ("millihartree", "Hz", "6.579683920502e9"),
        ("mhartree", "Hz", "6.579683920502e9"),
        ("hartree", "microgram", "4.8508702095432e-26"),
        ("microhartree", "wavenumber", "2.1947463136320e-7"),
        ("uhartree", "wavenumber", "2.1947463136320e-7"),
        ("kilohartree", "kiloeV", "27.211386245988"),
        ("hartree", "kiloeV", "0.027211386245988"),
        ("millihartree", "J/mol", "2625.4996394798"),
        ("hartree", "kJ/mol", "2625.4996394798"),
        ("hartree", "J", "4.3597447222071e-18"),
        ("1/m", "hartree", "4.556335252912e-08"),  # potentially another sigfig in 2018
    ],
)
def test_unit_conversion_nist_2018(from_unit, to_unit, expected, constants2018):
    converted = constants2018.conversion_factor(from_unit, to_unit)
    _assert_forward_conversion(expected, converted)


@pytest.mark.parametrize(
    "from_unit, to_unit, expected",
    [
        ("hartree", "kcal/mol", "627.509474"),
        ("hartree", "kJ/mol", "2625.499638"),
        ("bohr", "angstrom", "0.52917721067"),
        ("bohr", "nanometer", "5.2917721067e-2"),
        ("bohr", "cm", "5.2917721067e-9"),
        ("bohr", "miles", "3.288154743E-14"),
        ("amu", "kg", "1.660539040e-27"),
        ("amu", "g", "1.660539040e-24"),
        ("debye", "e * Bohr", "0.393430273"),
        # Derived atomic units (https://en.wikipedia.org/wiki/Hartree_atomic_units)
        ("au_pressure", "J / m^3", "2.94210e13"),
    ],
)
def test_unit_conversion_other_2014(from_unit, to_unit, expected, constants2014):
    converted = constants2014.conversion_factor(from_unit, to_unit)
    _assert_forward_conversion(expected, converted)

    converted = constants2014.conversion_factor(to_unit, from_unit)
    _assert_inverse_conversion(expected, converted)


@pytest.mark.parametrize(
    "from_unit, to_unit, expected",
    [
        ("hartree", "kcal/mol", "627.509474063"),  # 627.50947406305
        ("hartree", "kJ/mol", "2625.4996394798"),
        ("bohr", "angstrom", "0.529177210903"),
        ("bohr", "nanometer", "5.29177210903e-2"),
        ("bohr", "cm", "5.29177210903e-9"),
        ("bohr", "miles", "3.28815474444e-14"),
        ("amu", "kg", "1.66053906660e-27"),
        ("amu", "g", "1.66053906660e-24"),
        ("debye", "e * Bohr", "0.39343027"),  # 0.3934302694
        # Derived atomic units (https://en.wikipedia.org/wiki/Hartree_atomic_units)
        ("au_pressure", "J / m^3", "2.94210e13"),  # note precision loss due to absence in CODATA
    ],
)
def test_unit_conversion_other_2018(from_unit, to_unit, expected, constants2018):
    converted = constants2018.conversion_factor(from_unit, to_unit)
    _assert_forward_conversion(expected, converted)

    converted = constants2018.conversion_factor(to_unit, from_unit)
    _assert_inverse_conversion(expected, converted)


@pytest.mark.parametrize(
    "from_unit, to_unit, expected",
    [
        ("Bohr", "bohr", "1.0"),
        ("Angstrom", "Angstrom", "1.0"),
        ("coulomb", "statC", "2997924580"),
        # Derived atomic units (https://en.wikipedia.org/wiki/Hartree_atomic_units)
        ("au_1st_hyperpolarizability", "au_charge**3 * au_length**3 / au_energy**2", "1.0"),
        ("au_2nd_hyperpolarizability", "au_charge**4 * au_length**4 / au_energy**3", "1.0"),
        ("au_charge_density", "au_charge / au_length**3", "1.0"),
        ("au_current", "au_charge * au_energy / au_action", "1.0"),
        ("au_electric_dipole_moment", "au_charge * au_length", "1.0"),
        ("au_electric_field", "au_energy / (au_charge * au_length)", "1.0"),
        ("au_electric_field_gradient", "au_energy / (au_charge * au_length**2)", "1.0"),
        ("au_electric_polarizability", "au_charge**2 * au_length**2 / au_energy", "1.0"),
        ("au_electric_potential", "au_energy / au_charge", "1.0"),
        ("au_electric_quadrupole_moment", "au_charge * au_length**2", "1.0"),
        ("au_force", "au_energy / au_length", "1.0"),
        ("au_magnetic_dipole_moment", "au_charge * au_action / au_mass", "1.0"),
        ("au_magnetic_flux_density", "au_action / (au_charge * au_length**2)", "1.0"),
        ("au_magnetizability", "au_charge**2 * au_length**2 / au_mass", "1.0"),
        ("au_momentum", "au_action / au_length", "1.0"),
        ("au_permittivity", "au_charge**2 / (au_length * au_energy)", "1.0"),
        ("au_time", "au_action / au_energy", "1.0"),
        ("au_velocity", "au_length * au_energy / au_action", "1.0"),
    ],
)
def test_unit_conversion_other_exact(from_unit, to_unit, expected, contexts):
    converted = contexts.conversion_factor(from_unit, to_unit)
    _assert_forward_conversion(expected, converted)

    converted = contexts.conversion_factor(to_unit, from_unit)
    _assert_inverse_conversion(expected, converted)


def test_quantities_smoke():
    """
    Smoke test to ensure Quantities are correctly returned
    """
    assert 5 == qcelemental.constants.Quantity("5 kcal").magnitude


def test_speed_of_light(contexts):
    assert (
        pytest.approx(contexts.speed_of_light_in_vacuum * contexts.conversion_factor("m/s", "bohr/au_time"), 1e-8)
        == contexts.c_au
    )



from enum import Enum
from typing import TYPE_CHECKING, Any, Dict, Optional

import numpy as np
from pydantic import Field

from .basemodels import ProtoModel

if TYPE_CHECKING:
    from pydantic.typing import ReprArgs


# Encoders, to be deprecated
ndarray_encoder = {np.ndarray: lambda v: v.flatten().tolist()}


class Provenance(ProtoModel):
    """
    Provenance information.
    """

    creator: str = Field(..., description="The creator of the object.")
    version: Optional[str] = Field(None, description="The version of the creator.")
    routine: Optional[str] = Field(None, description="The routine of the creator.")

    class Config(ProtoModel.Config):
        canonical_repr = True
        extra = "allow"


class Model(ProtoModel):
    """
    The quantum chemistry model specification for a given operation to compute against
    """

    method: str = Field(  # type: ignore
        ..., description="The quantum chemistry method to evaluate (e.g., B3LYP, PBE, ...)."
    )
    basis: Optional[str] = Field(  # type: ignore
        None,
        description="The quantum chemistry basis set to evaluate (e.g., 6-31g, cc-pVDZ, ...). Can be ``None`` for "
        "methods without basis sets.",
    )

    # basis_spec: BasisSpec = None  # This should be exclusive with basis, but for now will be omitted

    class Config(ProtoModel.Config):
        canonical_repr = True
        extra = "allow"


class DriverEnum(str, Enum):
    """Allowed quantum chemistry driver values."""

    energy = "energy"
    gradient = "gradient"
    hessian = "hessian"
    properties = "properties"

    def derivative_int(self):
        egh = ["energy", "gradient", "hessian", "third", "fourth", "fifth"]
        if self == "properties":
            return 0
        else:
            return egh.index(self)


class ComputeError(ProtoModel):
    """A complete description of the error."""

    error_type: str = Field(  # type: ignore
        ...,  # Error enumeration not yet strict
        description="The type of error which was thrown. Restrict this field short classifiers e.g. 'input_error'.",
    )
    error_message: str = Field(  # type: ignore
        ...,
        description="Text associated with the thrown error, often the backtrace, but can contain additional "
        "information as well.",
    )
    extras: Optional[Dict[str, Any]] = Field(  # type: ignore
        None, description="Additional data to ship with the ComputeError object."
    )

    class Config:
        repr_style = ["error_type", "error_message"]

    def __repr_args__(self) -> "ReprArgs":
        return [("error_type", self.error_type), ("error_message", self.error_message)]


class FailedOperation(ProtoModel):
    """
    A record indicating that a given operation (compute, procedure, etc.) has failed and contains the reason and
    input data which generated the failure.

    """

    id: str = Field(  # type: ignore
        None,
        description="A unique identifier which links this FailedOperation, often of the same Id of the operation "
        "should it have been successful. This will often be set programmatically by a database such as "
        "Fractal.",
    )
    input_data: Any = Field(  # type: ignore
        None,
        description="The input data which was passed in that generated this failure. This should be the complete "
        "input which when attempted to be run, caused the operation to fail.",
    )
    success: bool = Field(  # type: ignore
        False,
        description="A boolean indicator that the operation failed consistent with the model of successful operations. "
        "Should always be False. Allows programmatic assessment of all operations regardless of if they failed or "
        "succeeded",
    )
    error: ComputeError = Field(  # type: ignore
        ...,
        description="A container which has details of the error that failed this operation. See the "
        ":class:`ComputeError` for more details.",
    )
    extras: Optional[Dict[str, Any]] = Field(  # type: ignore
        None,
        description="Additional information to bundle with this Failed Operation. Details which pertain specifically "
        "to a thrown error should be contained in the `error` field. See :class:`ComputeError` for details.",
    )

    def __repr_args__(self) -> "ReprArgs":
        return [("error", self.error)]


qcschema_input_default = "qcschema_input"
qcschema_output_default = "qcschema_output"
qcschema_optimization_input_default = "qcschema_optimization_input"
qcschema_optimization_output_default = "qcschema_optimization_output"
qcschema_molecule_default = "qcschema_molecule"



import os
import shutil
import sys
from typing import Union


def which_import(
    module: str,
    *,
    return_bool: bool = False,
    raise_error: bool = False,
    raise_msg: str = None,
    package: str = None,
    namespace_ok: bool = False,
) -> Union[bool, None, str]:
    """Tests to see if a Python module is available.

    Returns
    -------
    str or None
        By default, returns `__init__.py`-like path if `module` found or `None` if not.
        For namespace packages and if `namespace_ok=True`, returns the list of pieces locations if `module` found or `None` if not.
    bool
        When `return_bool=True`, returns whether or not found.
        Namespace packages only `True` if `namespace_ok=True`.

    Raises
    ------
    ModuleNotFoundError
        When `raise_error=True` and module not found. Raises generic message plus any `raise_msg`.

    """
    from importlib import util

    try:
        module_spec = util.find_spec(module, package=package)
    except ModuleNotFoundError:
        module_spec = None

    # module_spec.origin is 'namespace' for py36, None for >=py37
    namespace_package = module_spec is not None and module_spec.origin in [None, "namespace"]

    if (module_spec is None) or (namespace_package and not namespace_ok):
        if raise_error:
            raise ModuleNotFoundError(
                f"Python module '{module}' not found in envvar PYTHONPATH.{' ' + raise_msg if raise_msg else ''}"
            )
        elif return_bool:
            return False
        else:
            return None
    else:
        if return_bool:
            return True
        else:
            if namespace_package:
                return module_spec.submodule_search_locations
            else:
                return module_spec.origin


def which(
    command: str, *, return_bool: bool = False, raise_error: bool = False, raise_msg: str = None, env: str = None
) -> Union[bool, None, str]:
    """Test to see if a command is available.

    Returns
    -------
    str or None
        By default, returns command path if command found or `None` if not.
        Environment is $PATH or `os.pathsep`-separated `env`, less any None values.
    bool
        When `return_bool=True`, returns whether or not found.

    Raises
    ------
    ModuleNotFoundError
        When `raises_error=True` and command not found. Raises generic message plus any `raise_msg`.

    """
    if env is None:
        lenv = {"PATH": os.pathsep + os.environ.get("PATH", "") + os.path.dirname(sys.executable)}
    else:
        lenv = {"PATH": os.pathsep.join([os.path.abspath(x) for x in env.split(os.pathsep) if x != ""])}
    lenv = {k: v for k, v in lenv.items() if v is not None}

    ans = shutil.which(command, mode=os.F_OK | os.X_OK, path=lenv["PATH"])

    if raise_error and ans is None:
        raise ModuleNotFoundError(
            f"Command '{command}' not found in envvar PATH.{' ' + raise_msg if raise_msg else ''}"
        )

    if return_bool:
        return bool(ans)
    else:
        return ans


def safe_version(*args, **kwargs) -> str:
    """
    Package resources is a very slow load
    """
    import pkg_resources

    return pkg_resources.safe_version(*args, **kwargs)


def parse_version(*args, **kwargs):
    """
    Package resources is a very slow load
    """
    import pkg_resources

    return pkg_resources.parse_version(*args, **kwargs)



from decimal import Decimal

import numpy as np
import pytest

import qcelemental as qcel

_arrs = {
    "a1234_14": np.arange(4),
    "blip14": np.arange(4) + [0.0, 0.02, 0.005, 0.02],
    "nblip14": -np.arange(4) + [0.0, 0.02, 0.005, 0.02],
    "a1234_22": np.arange(4).reshape((2, 2)),
    "blip22": (np.arange(4) + [0.0, 0.02, 0.005, 0.02]).reshape((2, 2)),
    "iblip14": np.arange(4) + [0, 1, 0, 1],
    "iblip22": (np.arange(4) + [0, 1, 0, 1]).reshape((2, 2)),
}

_dcts = {
    "ell": {"a": _arrs["a1234_14"], "b": {"ba": _arrs["a1234_14"], "bb": _arrs["a1234_22"]}},
    "elll": {"a": _arrs["a1234_14"], "b": {"ba": _arrs["a1234_14"], "bb": _arrs["a1234_22"], "bc": 4}},
    "ellnone": {"a": _arrs["a1234_14"], "b": {"ba": _arrs["a1234_14"], "bb": _arrs["a1234_22"], "bc": None}},
    "ellshort": {"a": np.arange(3), "b": {"ba": _arrs["a1234_14"], "bb": _arrs["a1234_22"]}},
    "blipell": {"a": _arrs["blip14"], "b": {"ba": _arrs["a1234_14"], "bb": _arrs["blip22"]}},
    "nnell": {"a": -_arrs["a1234_14"], "b": {"ba": _arrs["a1234_14"], "bb": -_arrs["a1234_22"]}},
}

_pass_message = "\t{:.<66}PASSED"


@pytest.mark.parametrize(
    "ref,cpd,tol,kw,boool,msg",
    [
        (2.0, 2.02, 1.0e-1, {"label": "asdf"}, True, (None, _pass_message.format("asdf"))),
        (
            2.0,
            2.02,
            1.0e-2,
            {"label": "asdf"},
            False,
            (None, "asdf: computed value (2.0200) does not match (2.0000) to atol=0.01 by difference (0.0200)."),
        ),
        (2.0, Decimal("2.02"), 1.0e-1, {"label": "asdf"}, True, (None, _pass_message.format("asdf"))),
        (
            2.0,
            Decimal("2.02"),
            1.0e-2,
            {"label": "asdf"},
            False,
            (None, "asdf: computed value (2.0200) does not match (2.0000) to atol=0.01 by difference (0.0200)."),
        ),
        (
            2.0,
            2.000000002,
            1.0e-9,
            {},
            False,
            (
                None,
                "test_compare_values: computed value (2.00000000200) does not match (2.00000000000) to atol=1e-09 by difference (0.00000000200).",
            ),
        ),
        (
            2.0,
            2.05,
            1.0e-3,
            {},
            False,
            (
                None,
                "test_compare_values: computed value (2.05000) does not match (2.00000) to atol=0.001 by difference (0.05000).",
            ),
        ),
        (_arrs["a1234_14"], _arrs["blip14"], 1.0e-1, {}, True, (None, _pass_message.format("test_compare_values"))),
        (
            _arrs["a1234_14"],
            _arrs["nblip14"],
            1.0e-1,
            {"equal_phase": False},
            False,
            (
                None,
                """test_compare_values: computed value does not match to atol=0.1.
  Expected:
    [0. 1. 2. 3.]
  Observed:
    [ 0.    -0.98  -1.995 -2.98 ]
  Difference (passed elements are zeroed):
    [ 0.    -1.98  -3.995 -5.98 ]""",
            ),
        ),
        (
            _arrs["a1234_14"],
            _arrs["nblip14"],
            1.0e-1,
            {"equal_phase": True},
            True,
            (None, _pass_message.format("test_compare_values")),
        ),
        (
            _arrs["a1234_14"],
            _arrs["blip14"],
            1.0e-2,
            {},
            False,
            (
                None,
                """test_compare_values: computed value does not match to atol=0.01.
  Expected:
    [0. 1. 2. 3.]
  Observed:
    [0.    1.02  2.005 3.02 ]
  Difference (passed elements are zeroed):
    [0.   0.02 0.   0.02]""",
            ),
        ),
        (_arrs["a1234_22"], _arrs["blip22"], 1.0e-1, {}, True, (None, _pass_message.format("test_compare_values"))),
        (
            _arrs["a1234_22"],
            _arrs["blip22"],
            1.0e-2,
            {},
            False,
            (
                None,
                """test_compare_values: computed value does not match to atol=0.01.
  Expected:
    [[0. 1.]
     [2. 3.]]
  Observed:
    [[0.    1.02 ]
     [2.005 3.02 ]]
  Difference (passed elements are zeroed):
    [[0.   0.02]
     [0.   0.02]]""",
            ),
        ),
        (
            _arrs["a1234_22"],
            _arrs["blip14"],
            1.0e-2,
            {},
            False,
            (None, """test_compare_values: computed shape ((4,)) does not match ((2, 2))."""),
        ),
        (None, None, 1.0e-4, {"passnone": True}, True, (None, _pass_message.format("test_compare_values"))),
        (
            None,
            None,
            1.0e-4,
            {},
            False,
            (
                None,
                """test_compare_values: computed value (nan) does not match (nan) to atol=0.0001 by difference (nan).""",
            ),
        ),
    ],
)
def test_compare_values(ref, cpd, tol, kw, boool, msg):
    res, mstr = qcel.testing.compare_values(ref, cpd, **kw, atol=tol, return_message=True)
    assert res is boool
    assert mstr.strip() == msg[1].strip()


@pytest.mark.parametrize(
    "ref,cpd,kw,boool,msg",
    [
        (2, 2, {"label": "asdf"}, True, (None, _pass_message.format("asdf"))),
        (2, -2, {"label": "asdf"}, False, (None, "asdf: computed value (-2) does not match (2) by difference (-4).")),
        (2, Decimal("2"), {"label": "asdf"}, True, (None, _pass_message.format("asdf"))),
        (
            2,
            Decimal("3"),
            {"label": "asdf"},
            False,
            (None, "asdf: computed value (3) does not match (2) by difference (1)."),
        ),
        (True, True, {}, True, (None, _pass_message.format("test_compare"))),
        (False, False, {}, True, (None, _pass_message.format("test_compare"))),
        (None, None, {}, True, (None, _pass_message.format("test_compare"))),
        (
            True,
            None,
            {},
            False,
            (None, "test_compare: computed value (None) does not match (True) by difference ((n/a))."),
        ),
        (
            True,
            False,
            {},
            False,
            (None, "test_compare: computed value (False) does not match (True) by difference ((n/a))."),
        ),
        (
            False,
            None,
            {},
            False,
            (None, "test_compare: computed value (None) does not match (False) by difference ((n/a))."),
        ),
        (
            False,
            True,
            {},
            False,
            (None, "test_compare: computed value (True) does not match (False) by difference ((n/a))."),
        ),
        (
            None,
            False,
            {},
            False,
            (None, "test_compare: computed value (False) does not match (None) by difference ((n/a))."),
        ),
        (
            None,
            True,
            {},
            False,
            (None, "test_compare: computed value (True) does not match (None) by difference ((n/a))."),
        ),
        ("cat", "cat", {}, True, (None, _pass_message.format("test_compare"))),
        (
            "cat",
            "mouse",
            {},
            False,
            (None, "test_compare: computed value (mouse) does not match (cat) by difference ((n/a))."),
        ),
        ("cat", "cat", {"equal_phase": True}, True, (None, _pass_message.format("test_compare"))),
        (
            "cat",
            "mouse",
            {"equal_phase": True},
            False,
            (None, "test_compare: computed value (mouse) does not match (cat) by difference ((n/a))."),
        ),
        (_arrs["a1234_14"], _arrs["a1234_14"], {}, True, (None, _pass_message.format("test_compare"))),
        (
            _arrs["a1234_14"],
            _arrs["iblip14"],
            {},
            False,
            (
                None,
                """ test_compare: computed value does not match.
  Expected:
    [0 1 2 3]
  Observed:
    [0 2 2 4]
  Difference:
    [0 1 0 1]""",
            ),
        ),
        (_arrs["a1234_22"], _arrs["a1234_22"], {}, True, (None, _pass_message.format("test_compare"))),
        (
            _arrs["a1234_22"],
            _arrs["iblip22"],
            {},
            False,
            (
                None,
                """test_compare: computed value does not match.
  Expected:
    [[0 1]
     [2 3]]
  Observed:
    [[0 2]
     [2 4]]
  Difference:
    [[0 1]
     [0 1]]""",
            ),
        ),
        (
            _arrs["a1234_22"],
            _arrs["iblip14"],
            {},
            False,
            (None, """test_compare: computed shape ((4,)) does not match ((2, 2))."""),
        ),
        (1, np.ones(1), {}, False, (None, """test_compare: computed shape ((1,)) does not match (()).""")),
        (1, np.array(1), {}, True, (None, _pass_message.format("test_compare"))),
        (1, np.array([1]), {}, False, (None, """test_compare: computed shape ((1,)) does not match (()).""")),
    ],
)
def test_compare(ref, cpd, kw, boool, msg):
    res, mstr = qcel.testing.compare(ref, cpd, **kw, return_message=True)
    assert res is boool
    assert mstr.strip() == msg[1].strip()


@pytest.mark.parametrize(
    "ref,cpd,kw,boool,msg",
    [
        (_dcts["ell"], _dcts["ell"], {}, True, (None, "")),
        (
            _dcts["elll"],
            _dcts["ell"],
            {},
            False,
            (
                None,
                """root.b
    Missing keys {'bc'}""",
            ),
        ),
        (
            _dcts["ell"],
            _dcts["elll"],
            {},
            False,
            (
                None,
                """root.b
    Found extra keys {'bc'}""",
            ),
        ),
        (
            _dcts["ellnone"],
            _dcts["elll"],
            {},
            False,
            (
                None,
                """root.b.bc
    'None' does not match.""",
            ),
        ),
        (
            _dcts["ell"],
            _dcts["ellshort"],
            {},
            False,
            (
                None,
                """root.a
    Arrays differ.\t_compare_recursive: computed shape ((3,)) does not match ((4,)).""",
            ),
        ),
        (
            qcel.util.unnp(_dcts["ell"]),
            qcel.util.unnp(_dcts["ellshort"]),
            {},
            False,
            (
                None,
                """root.a
    Iterable lengths did not match""",
            ),
        ),
        (
            _dcts["ell"],
            _dcts["blipell"],
            {},
            False,
            (
                None,
                """root.a
    Arrays differ.\t_compare_recursive: computed value does not match.
  Expected:
    [0 1 2 3]
  Observed:
    [0.    1.02  2.005 3.02 ]
  Difference:
    [0.    0.02  0.005 0.02 ]

root.b.bb
    Arrays differ.\t_compare_recursive: computed value does not match.
  Expected:
    [[0 1]
     [2 3]]
  Observed:
    [[0.    1.02 ]
     [2.005 3.02 ]]
  Difference:
    [[0.    0.02 ]
     [0.005 0.02 ]]""",
            ),
        ),
        (
            {"a": [1, 2], "b": {"ba": [1, 2, 3], "bb": [1, 2, 3]}},
            {"a": [1, 3], "b": {"ba": [1, 4, 4], "bb": [1, 2, 4]}},
            {},
            False,
            (
                None,
                """root.a.1
    Value 2 did not match 3.
root.b.ba.1
    Value 2 did not match 4.
root.b.ba.2
    Value 3 did not match 4.
root.b.bb.2
    Value 3 did not match 4.""",
            ),
        ),
        (_dcts["elll"], _dcts["ell"], {"forgive": ["root.b"]}, True, (None, "")),
        (_dcts["ell"], _dcts["elll"], {"forgive": ["b"]}, True, (None, "")),
        (_dcts["ellnone"], _dcts["elll"], {"forgive": ["b"]}, True, (None, "")),
        (_dcts["ellnone"], _dcts["elll"], {"forgive": ["b.bc"]}, True, (None, "")),
        (
            _dcts["ell"],
            _dcts["blipell"],
            {"forgive": ["b"]},
            False,
            (
                None,
                """root.a
    Arrays differ.\t_compare_recursive: computed value does not match.
  Expected:
    [0 1 2 3]
  Observed:
    [0.    1.02  2.005 3.02 ]
  Difference:
    [0.    0.02  0.005 0.02 ]""",
            ),
        ),
        (
            {"a": [1, 2], "b": {"ba": [1, 2, 3], "bb": [1, 2, 3]}},
            {"a": [1, 3], "b": {"ba": [1, 4, 4], "bb": [1, 2, 4]}},
            {"forgive": ["root.a", "b.bb.2"]},
            False,
            (
                None,
                """root.b.ba.1
    Value 2 did not match 4.
root.b.ba.2
    Value 3 did not match 4.""",
            ),
        ),
        (
            _dcts["ell"],
            _dcts["nnell"],
            {},
            False,
            (
                None,
                """root.a
    Arrays differ.\t_compare_recursive: computed value does not match.
  Expected:
    [0 1 2 3]
  Observed:
    [ 0 -1 -2 -3]
  Difference:
    [ 0 -2 -4 -6]

root.b.bb
    Arrays differ.\t_compare_recursive: computed value does not match.
  Expected:
    [[0 1]
     [2 3]]
  Observed:
    [[ 0 -1]
     [-2 -3]]
  Difference:
    [[ 0 -2]
     [-4 -6]]
""",
            ),
        ),
        (
            _dcts["ell"],
            _dcts["nnell"],
            {"equal_phase": ["a"]},
            False,
            (
                None,
                """root.b.bb
    Arrays differ.\t_compare_recursive: computed value does not match.
  Expected:
    [[0 1]
     [2 3]]
  Observed:
    [[ 0 -1]
     [-2 -3]]
  Difference:
    [[ 0 -2]
     [-4 -6]]
""",
            ),
        ),
        (_dcts["ell"], _dcts["nnell"], {"forgive": ["a"], "equal_phase": ["b"]}, True, (None, "")),
        (
            _dcts["ell"],
            _dcts["nnell"],
            {"equal_phase": ["b.bb"]},
            False,
            (
                None,
                """root.a
    Arrays differ.\t_compare_recursive: computed value does not match.
  Expected:
    [0 1 2 3]
  Observed:
    [ 0 -1 -2 -3]
  Difference:
    [ 0 -2 -4 -6]
""",
            ),
        ),
        (_dcts["ell"], _dcts["nnell"], {"equal_phase": ["a", "b.bb"]}, True, (None, "")),
        (_dcts["ell"], _dcts["nnell"], {"equal_phase": True}, True, (None, "")),
    ],
)
def test_compare_recursive(ref, cpd, kw, boool, msg):
    res, mstr = qcel.testing.compare_recursive(ref, cpd, **kw, return_message=True)
    assert res is boool
    assert mstr.strip() == msg[1].strip()
